@@ -20,14 +20,12 @@ require 'fluent/msgpack_factory'
 module Fluent
   module Counter
     class BaseSocket < Coolio::TCPSocket
-      include Fluent::MessagePackFactory::Mixin
-
       def packed_write(data)
         write pack(data)
       end
 
       def on_read(data)
-        msgpack_unpacker.feed_each(data) do |d|
+        Fluent::MessagePackFactory.msgpack_unpacker.feed_each(data) do |d|
           on_message d
         end
       end
@@ -39,7 +37,7 @@ module Fluent
       private
 
       def pack(data)
-        msgpack_packer.pack(data)
+        Fluent::MessagePackFactory.msgpack_packer.pack(data)
       end
     end
   end
